/** \file listener.h
 * \brief TINE static listener documentation 
 *
 */
#ifndef LISTENER_H_
#define LISTENER_H_

#define ALLOWED_IDLE_TIME 300
typedef struct mexListenerStruct
{
  char devName[64];
  char devProperty[32];
  DTYPE dout;
  DTYPE din;
  BYTE *outbuf;
  BYTE *inbuf;
  int lid;
  int status;
  int idleTimeInSeconds;
  int isGlobal;
  double timestamp;
  struct mexListenerStruct *nxt;
} mexListener;

/* setDefaultListenerMode() */
/**
 * \brief Sets the polling mode for any static listeners created via a call to
 * alsnExecLink().
 *
 * The polling mode to be used in establishing an asynchronous static listener is
 * not a input parameter to the api call alsnExecLink().  Instead the polling mode
 * which is applied is determined by the current value of the default listener mode.
 * (CM_POLL upon startup).  In any case the CM_WAIT flag is applied the the poll mode
 * so that the synchronous call alsnExecLink() does not return until the initial call
 * has completed.
 * 
 * \param mode is the desired default polling mode, i.e. one of CM_POLL or CM_REFRESH
 * (possibly ORed with CM_CONNECT or CM_NETWORK for instance).  Setting the default 
 * mode to CM_SINGLE or CM_CANCEL makes no sense in the context of a static listener.
 *
 * \return The current default polling mode to be applied to a static listener.
 * 
 */
int setDefaultListenerMode(int mode);
int stopListener(mexListener *ml);
int isGlobal(char *devName);
mexListener *startListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,int pollrate);
int stopListener(mexListener *ml);
mexListener *getListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din);
int getDataFromListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,long rate);
int isGlobal(char *devName);
int isMetaCall(char *devName,char *devProperty);
int isQueryProperty(char *devProperty);
int getPropertyInfoFromListener(char *dev, char *prp, int *siz, int *fmt);

/* alsnExecLink() */
/**
 * \brief Executes a synchronous call within an asynchronous static listener.
 * 
 * Asynchronous/Synchronous data exchange. For query calls (for instance to obtain a property
 * list), calls to the local history server, or calls with a timeout parameter less than 50 
 * milliseconds, the call reverts to ExecLinkEx(), i.e. synchronous data transfer.  Otherwise,
 * the initial call starts an asynchronous static listener for the selected device name and
 * property, polling at the rate determined by the timeout parameter. 
 * The initial call waits for the first data set to be returned before it completes.  Subsequent
 * calls retrieve data from the listener's buffer and complete immediately.  If no subsequent
 * calls are made within a 5 minute interval, the asynchronous link is cancelled.
 *
 * \param devName is the full device name (/\<Context\>/\<Server\>/\<Device\>) of the device to contact.
 * For example: "/HERA/BPM/WL167". 
 * \param devProperty is the device property requested, for example "ORBIT.X"
 * \param dout is a pointer to the output data set, that is, the data set to be returned from the server 
 * to the client.  This should refer to a DTYPE object giving the data size, format, and tag (if any)
 * and reference to the data of the output data set.
 * \param din is a pointer to the input data set, that is, the data set to be sent to the server from the 
 * client.  This should refer to a DTYPE object giving the data size, format, and tag (if any)
 * and reference to the data of the input data set.
 * \param access is the data access mode.  This can be any combination of access codes ORed together (CA_READ,
 * CA_WRITE, CA_CONNECT, etc.).
 * \param timeout the timeout period in milliseconds for the call to complete.  When it is known a priori 
 * that a call takes a long time to complete than you can set this value accordingly.  This parameter
 * has a particular meaning when used in the static listener routines.  Namely, it is also used to
 * determine the polling interval for the asynchronous listener which is attached to the call.  You
 * can force the call to execute synchronously without starting a listener by passing a value less
 * than 50 for this parameter.
 *
 * \return 0 if successful, otherwise a TINE completion code which can be interpreted by examining
 * RPCLastErrorString.
 *
 * \see ExecLink(), ExecLinkEx(), DTYPE
 */
int alsnExecLink(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access,long timeout);

extern mexListener *mexListeners;
extern int nListeners;

#endif /* LISTENER_H_ */
