/** \file listener.c
 * \brief TINE static listener documentation 
 *
 */
#ifndef NO_DOXYGEN
#include "prolog.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <malloc.h>
#include <sys/timeb.h>
#include "rpctypes.h"
#include "errors.h"
#include "database.h"
#include "brdcast.h"
#include "listener.h"

mexListener *mexListeners = NULL;
int nListeners = 0;
int defaultPollingMode = CM_POLL;

void listenerCb(int id,int cc)
{
  mexListener *ml = (mexListener *)id;
  if (!ml->isGlobal)
  {
    if (ml->idleTimeInSeconds++ > ALLOWED_IDLE_TIME) stopListener(ml);
    ml->timestamp = getCurrentDataTimeStamp(ml->lid);
  }
  else
  {
    ml->timestamp = (double)NGlobalList[ml->lid]->time;
  }
  return;
}
mexListener *startListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,int pollrate)
{
  int lid,siz;
  mexListener *ml=NULL;

# ifdef USE_QUERY_INFO
  XPropertyQueryStruct xpqs;
  siz = 1;
  if (GetDevicePropertyInformationX(devName,devProperty,&xpqs,&siz) == 0)
  {
    if (dout->dArrayLength < xpqs.prpSize) dout->dArrayLength = xpqs.prpSize;
  }
# endif
  siz = dout->dArrayLength * fmtsizeof(dout->dFormat);
  if ((ml=(mexListener *)calloc(1,sizeof(mexListener))) == NULL) return NULL;
  if ((ml->outbuf=(BYTE *)calloc(siz,1)) == NULL) 
  {
    free(ml);
    return NULL;
  }
  ml->dout.dArrayLength = dout->dArrayLength;
  ml->dout.dFormat = dout->dFormat;
  memcpy(ml->dout.dTag,dout->dTag,8);
  ml->dout.data.vptr = ml->outbuf;
  if (din != NULL && (siz=din->dArrayLength * fmtsizeof(din->dFormat)) > 0)
  {
    if ((ml->inbuf=(BYTE *)calloc(siz,1)) == NULL) 
    {
      if (ml->outbuf != NULL) free(ml->outbuf);
      free(ml);
      return NULL;
    }
    memcpy(ml->din.data.vptr,din->data.vptr,siz);
    ml->din.dArrayLength = din->dArrayLength;
    ml->din.dFormat = din->dFormat;
    memcpy(ml->din.dTag,din->dTag,8);
  }
  else
  {
    ml->din.dFormat = CF_NULL;
    ml->din.dArrayLength = 0;
    ml->din.data.vptr = NULL;
    memset(ml->din.dTag,0,8);
  }
  strncpy(ml->devName,devName,64);
  strncpy(ml->devProperty,devProperty,32);

  if (isGlobal(devName))
  {
    lid = recvNetGlobalEx(devProperty,&ml->dout,listenerCb,(UINT32)ml);
    ml->isGlobal = -1;
  }
  else
  {
    lid = AttachLinkEx(devName,devProperty,&ml->dout,&ml->din,CA_READ,pollrate,listenerCb,defaultPollingMode|CM_WAIT,(UINT32)ml);
    ml->isGlobal = 0;
  }
  if (lid < 0)
  {
    if (ml->outbuf != NULL) free(ml->outbuf);
    if (ml->inbuf != NULL) free(ml->inbuf);
    free(ml);
    return NULL;
  }
  if (ml->dout.dArrayLength != 0)
  {
    memcpy(dout->data.vptr,ml->outbuf,dout->dArrayLength*fmtsizeof(dout->dFormat));
  }
  ml->lid = lid;
  ml->nxt = mexListeners;
  mexListeners = ml;
  return ml;
}
int stopListener(mexListener *ml)
{
  mexListener *m = NULL;
  if (ml == NULL) return non_existent_elem;
  Cancel(ml->lid);
  if (ml == mexListeners) mexListeners = ml->nxt;
  else
  {
    for (m=mexListeners; m != NULL && m->nxt != ml; m = m->nxt);
    if (m != NULL) m->nxt = ml->nxt;
  }
  if (ml->outbuf != NULL) free(ml->outbuf);
  if (ml->inbuf != NULL) free(ml->inbuf);
  free(ml);
  return 0;
}
mexListener *getListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din)
{
  mexListener *ml=NULL;
  UINT32 dsiz;
  for (ml=mexListeners; ml != NULL; ml = ml->nxt)
  {
    if (strncmp(devName,ml->devName,64)) continue;
    if (strlen(devProperty) > 0 && strncmp(devProperty,ml->devProperty,32)) continue;
    if (dout != NULL)
    {
      if (dout->dFormat != ml->dout.dFormat) continue;
      if (dout->dArrayLength != ml->dout.dArrayLength) continue;
    }
    if (din != NULL && ml->din.dArrayLength != 0)
    {
      if (din->dFormat != ml->din.dFormat) continue;
      if (din->dArrayLength != ml->din.dArrayLength) continue;
      dsiz = ml->din.dArrayLength*fmtsizeof(ml->din.dFormat);
      if (memcmp(din->data.vptr,ml->inbuf,dsiz)) continue;
    }
    break;
  }
  return ml;  
}
int getDataFromListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,long rate)
{
  mexListener *ml;
  UINT32 siz;

  if (devName == NULL || devProperty == NULL || dout == NULL) return argument_list_error;
  if ( (ml=getListener(devName,devProperty,dout,din)) == NULL && 
       (ml=startListener(devName,devProperty,dout,din,rate)) == NULL )
  {
    return un_allocated;
  }
  if (dout->dFormat != ml->dout.dFormat) return illegal_format;
  ml->idleTimeInSeconds = 0;
  dout->dTimeStamp = ml->timestamp;
  siz = MIN(dout->dArrayLength,ml->dout.dArrayLength)*fmtsizeof(dout->dFormat);
  if (dout->data.vptr != ml->outbuf) memcpy(dout->data.vptr,ml->outbuf,siz);
  return ml->status;
}
int getPropertyInfoFromListener(char *dev, char *prp, int *siz, int *fmt)
{
  mexListener *ml;

  if ((ml=getListener(dev,prp,NULL,NULL)) != NULL)
  {
    if (siz != NULL) *siz = ml->dout.dArrayLength;
    if (fmt != NULL) *fmt = ml->dout.dFormat;
    return 0;
  }
  return un_allocated;
}
int isGlobal(char *devName)
{
  if (strstr(devName,"GLOBQRY") != NULL) return -1;
  if (strstr(devName,GLOBALSRV) != NULL) return -1;
  return 0;
}
int isMetaCall(char *devName,char *devProperty)
{
  if (strstr(devName,"HISTORY") != NULL) return -1;
  if (strstr(devProperty,".HIST") != NULL) return -1;
  if (strstr(devProperty,".EGU") != NULL) return -1;
  if (strstr(devProperty,".NAM") != NULL) return -1;
  if (strstr(devProperty,".DESC") != NULL) return -1;
  return 0;
}
int isQueryProperty(char *devProperty)
{
  if (strstr(devProperty,"PROPS") != NULL) return -1;
  if (strstr(devProperty,"PROPERTIES") != NULL) return -1;
  if (strstr(devProperty,"DEVICES") != NULL) return -1;
  if (strstr(devProperty,"TAGS") != NULL) return -1;
  if (strstr(devProperty,"CONTEXTS") != NULL) return -1;
  if (strstr(devProperty,"SUBSYSTEMS") != NULL) return -1;
  if (strstr(devProperty,"USERS") != NULL) return -1;
  if (strstr(devProperty,"IPNETS") != NULL) return -1;
  if (strstr(devProperty,"CLIENTS") != NULL) return -1;
  if (strstr(devProperty,"CONTRACTS") != NULL) return -1;
  if (strstr(devProperty,"HISTORIES") != NULL) return -1;
  if (strstr(devProperty,"LOGFILE") != NULL) return -1;
  if (strstr(devProperty,"LOGDEPTH") != NULL) return -1;
  if (strstr(devProperty,"STRUCTFORMAT") != NULL) return -1;
  if (strncmp(devProperty,"SRV",3) == 0) return -1;
  if (strstr(devProperty,"APPDATE") != NULL) return -1;
  if (strstr(devProperty,"APPVERSION") != NULL) return -1;
  return 0;
}
#endif /* NO_DOXYGEN */
int setDefaultListenerMode(int mode)
{
  if (mode > 0) defaultPollingMode = mode;
  return defaultPollingMode;
}
int alsnExecLink(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access,long timeout)
{
  int cc,retry=2,len;
  PropertyQueryStruct svrProp;
  DTYPE lcldin, *dinptr;

  if (din != NULL) 
  {
    dinptr = din;
  }
  else
  {
    lcldin.dArrayLength = 0;
    lcldin.dFormat = CF_NULL;
    lcldin.data.vptr = NULL;
    dinptr = &lcldin;
  }
  if (isGlobal(devName))
  {
    if (dout == NULL) return argument_list_error;
    if (!strncmp(devProperty,"NPROPS",6))
    {
      dout->data.sptr[0] = 1;
      return 0;
    }
    if (!strncmp(devProperty,"PROPS",5))
    {
      if ((len=dinptr->dArrayLength) > 31) len = 31; 
      strncpy(svrProp.prpName,dinptr->data.cptr,len); svrProp.prpName[len] = 0;
      svrProp.prpSize = 1;
      if (strstr(dinptr->data.cptr,"OPTICS") || strstr(dinptr->data.cptr,"STATE"))
      {
        svrProp.prpFormat = BFMT(CF_NAME16);
      }
      else
      {
        svrProp.prpFormat = BFMT(CF_FLOAT);
      } 
      memcpy(dout->data.vptr,&svrProp,sizeof(PropertyQueryStruct));
      return 0;
    }
  }
  if (!(access&CA_WRITE) && !isQueryProperty(devProperty) && 
      !isMetaCall(devName,devProperty) && timeout >= 50)
  {
    return getDataFromListener(devName,devProperty,dout,dinptr,timeout);
  }
  while (retry-- && timeout < 1000)
  {
    if (timeout < 10) timeout *= 1000;
    if ((cc=ExecLinkEx(devName,devProperty,dout,dinptr,access,(UINT16)timeout)) == 0) break;
  }
  return cc;
}
